@@ -50,6 +50,7 @@ module Hookup
   CommandReply(..)
 
   -- * SSL Information
+  , getClientCertificate
   , getPeerCertificate
   , getPeerCertFingerprintSha1
   , getPeerCertFingerprintSha256
@@ -62,6 +63,7 @@ module Hookup
 import           Control.Concurrent
 import           Control.Exception
 import           Control.Monad
+import           System.IO.Error (isDoesNotExistError, ioeGetErrorString)
 import           Data.ByteString (ByteString)
 import qualified Data.ByteString as B
 import qualified Data.ByteString.Char8 as B8
@@ -128,7 +130,7 @@ data TlsParams = TlsParams
 -- | Type for errors that can be thrown by this package.
 data ConnectionFailure
   -- | Failure during 'getAddrInfo' resolving remote host
-  = HostnameResolutionFailure IOError
+  = HostnameResolutionFailure HostName String
   -- | Failure during 'connect' to remote host
   | ConnectionFailure [IOError]
   -- | Failure during 'recvLine'
@@ -152,8 +154,8 @@ instance Exception ConnectionFailure where
   displayException (ConnectionFailure xs) =
     "connection attempt failed due to: " ++
       intercalate ", " (map displayException xs)
-  displayException (HostnameResolutionFailure x) =
-    "hostname resolution failed: " ++ displayException x
+  displayException (HostnameResolutionFailure h s) =
+    "hostname resolution failed (" ++ h ++ "): "  ++ s
   displayException SocksAuthenticationError =
     "SOCKS authentication method rejected"
   displayException SocksProtocolError =
@@ -262,7 +264,10 @@ openSocket' family h p =
      res <- try (Socket.getAddrInfo (Just hints) (Just h) (Just (show p)))
      case res of
        Right ais -> attemptConnections [] ais
-       Left  ioe -> throwIO (HostnameResolutionFailure ioe)
+       Left  ioe
+         | isDoesNotExistError ioe ->
+             throwIO (HostnameResolutionFailure h (ioeGetErrorString ioe))
+         | otherwise -> throwIO ioe -- unexpected
 
 
 -- | Try establishing a connection to the services indicated by
@@ -301,7 +306,7 @@ socket' ai =
 -- Generalization of Socket
 ------------------------------------------------------------------------
 
-data NetworkHandle = SSL SSL | Socket Socket
+data NetworkHandle = SSL (Maybe X509) SSL | Socket Socket
 
 
 openNetworkHandle ::
@@ -311,22 +316,24 @@ openNetworkHandle ::
 openNetworkHandle params mkSocket =
   case cpTls params of
     Nothing  -> Socket <$> mkSocket
-    Just tls -> SSL <$> startTls tls (cpHost params) mkSocket
+    Just tls ->
+        do (clientCert, ssl) <- startTls tls (cpHost params) mkSocket
+           pure (SSL clientCert ssl)
 
 
 closeNetworkHandle :: NetworkHandle -> IO ()
 closeNetworkHandle (Socket s) = Socket.close s
-closeNetworkHandle (SSL s) =
+closeNetworkHandle (SSL _ s) =
   do SSL.shutdown s SSL.Unidirectional
      traverse_ Socket.close (SSL.sslSocket s)
 
 networkSend :: NetworkHandle -> ByteString -> IO ()
 networkSend (Socket s) = SocketB.sendAll s
-networkSend (SSL    s) = SSL.write       s
+networkSend (SSL  _ s) = SSL.write       s
 
 networkRecv :: NetworkHandle -> Int -> IO ByteString
 networkRecv (Socket s) = SocketB.recv s
-networkRecv (SSL    s) = SSL.read     s
+networkRecv (SSL  _ s) = SSL.read     s
 
 
 ------------------------------------------------------------------------
@@ -468,7 +475,7 @@ startTls ::
   TlsParams {- ^ connection params      -} ->
   String    {- ^ hostname               -} ->
   IO Socket {- ^ socket creation action -} ->
-  IO SSL    {- ^ connected TLS          -}
+  IO (Maybe X509, SSL) {- ^ (client certificate, connected TLS) -}
 startTls tp hostname mkSocket = SSL.withOpenSSL $
   do ctx <- SSL.context
 
@@ -480,8 +487,8 @@ startTls tp hostname mkSocket = SSL.withOpenSSL $
      SSL.contextRemoveOption        ctx SSL.SSL_OP_DONT_INSERT_EMPTY_FRAGMENTS
 
      -- configure certificates
-     setupCaCertificates ctx          (tpServerCertificate tp)
-     traverse_ (setupCertificate ctx) (tpClientCertificate tp)
+     setupCaCertificates ctx (tpServerCertificate tp)
+     clientCert <- traverse (setupCertificate ctx) (tpClientCertificate tp)
      traverse_ (setupPrivateKey  ctx) (tpClientPrivateKey  tp)
 
      -- add socket to context
@@ -494,7 +501,7 @@ startTls tp hostname mkSocket = SSL.withOpenSSL $
 
      SSL.connect ssl
 
-     return ssl
+     return (clientCert, ssl)
 
 
 setupCaCertificates :: SSLContext -> Maybe FilePath -> IO ()
@@ -504,11 +511,11 @@ setupCaCertificates ctx mbPath =
     Just path -> SSL.contextSetCAFile ctx path
 
 
-setupCertificate :: SSLContext -> FilePath -> IO ()
-setupCertificate ctx path
-  =   SSL.contextSetCertificate ctx
-  =<< PEM.readX509 -- EX
-  =<< readFile path
+setupCertificate :: SSLContext -> FilePath -> IO X509
+setupCertificate ctx path =
+  do x509 <- PEM.readX509 =<< readFile path -- EX
+     SSL.contextSetCertificate ctx x509
+     pure x509
 
 
 setupPrivateKey :: SSLContext -> FilePath -> IO ()
@@ -532,7 +539,14 @@ getPeerCertificate :: Connection -> IO (Maybe X509.X509)
 getPeerCertificate (Connection _ h) =
   case h of
     Socket{} -> return Nothing
-    SSL ssl  -> SSL.getPeerCertificate ssl
+    SSL _ ssl -> SSL.getPeerCertificate ssl
+
+-- | Get peer certificate if one exists.
+getClientCertificate :: Connection -> Maybe X509.X509
+getClientCertificate (Connection _ h) =
+  case h of
+    Socket{} -> Nothing
+    SSL c _  -> c
 
 getPeerCertFingerprintSha1 :: Connection -> IO (Maybe ByteString)
 getPeerCertFingerprintSha1 = getPeerCertFingerprint "sha1"
